from __future__ import annotations
import numpy as np

# import pandas as pd

# no-teg is quick and dirty, in the future could evolve to a full scale tournament API


# game could be reduced to a pd df
# each game just a new row with their specs


# game interface
class Game:
    """Initialize new game"""

    def __init__(self):
        self.name = None
        self.rec_players = None
        self.rec_tourney = None
        self.labels = []

    def set_name(self, name: str):
        self.name = name

    def set_rec_players(self, rec_players: int):
        self.rec_players = rec_players

    def set_rec_tourney(self, rec_tourney):
        self.rec_tourney = rec_tourney

    def set_labels(self, labels: list[str]):
        self.labels = labels

    def get_labels(self):
        return self.labels


class FIFA(Game):
    def __init__(self):
        self.name = "FIFA"
        self.rec_players = 8  # useless?
        self.rec_tourney = "Single_Elimination"
        self.labels = []


# interface
class Tourney:
    def __init__(self, game):
        self.game = game
        self.players = []
        self.matchups = {}

    def add_player(self, player: Player / Team):
        self.players.append(player)

    def add_players(self, players: list[Player / Team]):
        self.players = players

    def randomize_matchups(self):
        np.random.shuffle(self.players)

    def start(self, additional_stats=None):
        """Set all matchups"""
        pass

    def input_result(self):
        pass

    def get_matchups(self):
        return self.matchups

    # can take in a round to print the round
    def print_matchups(self):
        pass

    def print_results(self):
        pass


class Single_Elimination(Tourney):
    # ex matchups 4 person SE {
    #    1:{Home: "Aaron", Away: "Xandra", Next: 3, Home_Score: 2, Away_Score: 1},
    #    2:{Home: "Tiffany", Away: "Lucas", Next: 3, Home_Score: 4, Away_Score: 0},
    #    3:{Home: "Aaron", Away: "Lucas", Next: None, Home_Score: None, Away_Score: None}
    #   }
    def start(self):
        extra_labels = self.game.get_labels()
        # print("EXTRA LABELS: " + str(extra_labels))
        # set entire bracket
        # assume number of participants is a power of 2
        self.started = True
        # total_matches = len(self.players) - 1
        num_rounds = np.ceil(np.log2(len(self.players)))
        # num_byes = num_rounds**2 - len(self.players)
        matchup_counter = 1
        p1 = 0
        p2 = 1
        round_matches = len(self.players) // 2  # players power of 2
        round = 1
        updating_round_matches = 0
        while round_matches > 0:
            updating_round_matches += round_matches
            for i in range(round_matches):
                if round == 1:
                    home, away = self.players[p1].get_name(), self.players[p2].get_name()
                else:
                    home, away = None, None
                if round != num_rounds:
                    next = updating_round_matches + (i // 2) + 1
                else:
                    next = None
                self.matchups[matchup_counter] = {
                    "Home": home,
                    "Away": away,
                    "Next": next,
                    "Home_Score": None,
                    "Away_Score": None,
                }
                for stat in extra_labels:
                    self.matchups[matchup_counter][stat] = None
                matchup_counter += 1
                p1 += 2
                p2 += 2
            round_matches = round_matches // 2
            round += 1

    def input_result(self, matchup_id, away_score, home_score, extra_stats=[]):
        self.matchups[matchup_id]["Home_Score"] = home_score
        self.matchups[matchup_id]["Away_Score"] = away_score
        home = self.matchups[matchup_id]["Home"]
        away = self.matchups[matchup_id]["Away"]
        if home_score > away_score:
            winner = home
        elif away_score > home_score:
            winner = away
        else:
            winner = "Tie"
            print("This format does not support ties")  # dont let teis be input
            return False
        next = self.matchups[matchup_id]["Next"]
        if next is not None:
            if matchup_id % 2 == 0:
                self.matchups[next]["Home"] = winner
            else:
                self.matchups[next]["Away"] = winner
        extra_labels = self.game.get_labels()
        if len(extra_stats) == len(extra_labels):
            for i in range(len(extra_stats)):
                self.matchups[matchup_id][extra_labels[i]] = extra_stats[i]

    def print_matchups(self):
        for i in range(len(self.matchups)):
            matchup_id = i + 1
            home = self.matchups[matchup_id]["Home"]
            away = self.matchups[matchup_id]["Away"]
            if home is not None and away is not None:
                print("{:d}: {:s} (A) vs {:s} (H)".format(matchup_id, away, home))

    def print_results(self):
        for i in range(len(self.matchups)):
            matchup_id = i + 1
            home = self.matchups[matchup_id]["Home"]
            away = self.matchups[matchup_id]["Away"]
            home_score = self.matchups[matchup_id]["Home_Score"]
            away_score = self.matchups[matchup_id]["Away_Score"]
            if home is not None and away is not None and home_score is not None and away_score is not None:
                print("{:d}: {:s} ({:d}) vs {:s} ({:d})".format(matchup_id, away, away_score, home, home_score))


class Player:
    def __init__(self, name):
        self.name = name
        self.age = None

    def set_age(self, age):  # DOB?
        self.age = age

    def get_name(self):
        return self.name


class Team:
    def __init__(self, name: str):
        self.name = name
        self.players = []

    def add_player(self, player: Player):
        self.players.append(player)

    def add_players(self, players: list[Player]):
        self.players = players

    def get_name(self):
        return self.name

    def get_players(self):
        return self.players


def example1():
    MyTourney = Single_Elimination(FIFA())
    p1 = Player("Aaron")
    p2 = Player("Xandra")
    p3 = Player("Lucas")
    p4 = Player("Tiffany")
    MyTourney.add_players([p1, p2, p3, p4])
    MyTourney.randomize_matchups()
    MyTourney.start()
    MyTourney.print_matchups()
    MyTourney.input_result(1, 2, 3)
    MyTourney.input_result(2, 10, 1)
    MyTourney.input_result(3, 4, 1)
    MyTourney.print_results()



from no_teg import *
from unittest.mock import patch


### UNIT TESTS ###

TestTourney1 = Single_Elimination(FIFA())
p1 = Player("Aaron")
p2 = Player("Xandra")
p3 = Player("Lucas")
p4 = Player("Tiffany")


def test_Player_get_name():
    p1 = Player("Aaron")
    assert p1.get_name() == "Aaron"


def test_Player_set_age():
    p1.set_age(22)
    assert p1.age == 22


def test_Tourney_add_players():
    TestTourney1.add_players([p1, p2, p3, p4])
    assert TestTourney1.players == [p1, p2, p3, p4]


# use a real error formula to check randomization
def test_Tourney_randomize_matchups():
    times_first = {p1: 0, p2: 0, p3: 0, p4: 0}
    times_in_order = 0  # 1/24 chance
    trials = 2000
    for i in range(trials):
        TestTourney1.randomize_matchups()
        times_first[TestTourney1.players[0]] += 1
        if TestTourney1.players == [p1, p2, p3, p4]:
            times_in_order += 1
    for player in TestTourney1.players:
        expected = trials / 4
        error = 0.2
        assert times_first[player] > expected - (expected * error) and times_first[player] < expected + (
            expected * error
        )
    expected = trials / 24  # 4!
    error = 0.2
    assert times_in_order > expected - (expected * error) and times_in_order < expected + (expected * error)


TestTourneyOrdered = Single_Elimination(FIFA())
p5 = Player("Aaron")
p6 = Player("Xandra")
p7 = Player("Lucas")
p8 = Player("Tiffany")


def test_SE_start():
    TestTourneyOrdered.add_players([p1, p2, p3, p4])
    TestTourneyOrdered.start()
    assert TestTourneyOrdered.matchups[1]["Away"] == "Xandra"
    assert TestTourneyOrdered.matchups[1]["Home"] == "Aaron"
    assert TestTourneyOrdered.matchups[1]["Home_Score"] == None
    assert TestTourneyOrdered.matchups[1]["Away_Score"] == None
    assert TestTourneyOrdered.matchups[1]["Next"] == 3
    assert TestTourneyOrdered.matchups[2]["Away"] == "Tiffany"
    assert TestTourneyOrdered.matchups[2]["Home"] == "Lucas"
    assert TestTourneyOrdered.matchups[3]["Home"] == None
    assert TestTourneyOrdered.matchups[3]["Away"] == None


def test_SE_get_matchups():
    assert TestTourneyOrdered.get_matchups() == TestTourneyOrdered.matchups


def test_SE_input_result():
    TestTourneyOrdered.input_result(1, 0, 1)
    assert TestTourneyOrdered.matchups[1]["Away_Score"] == 0
    assert TestTourneyOrdered.matchups[1]["Home_Score"] == 1
    assert TestTourneyOrdered.matchups[3]["Away"] == "Aaron"
    TestTourneyOrdered.input_result(2, 4, 2)
    assert TestTourneyOrdered.matchups[2]["Away_Score"] == 4
    assert TestTourneyOrdered.matchups[2]["Home_Score"] == 2
    assert TestTourneyOrdered.matchups[3]["Home"] == "Tiffany"
    assert TestTourneyOrdered.matchups[3]["Next"] == None
    assert TestTourneyOrdered.input_result(3, 1, 1) == False


myGame = Game()


def test_game():
    assert myGame.name == None
    assert myGame.rec_players == None
    assert myGame.rec_tourney == None
    assert myGame.labels == []


def test_game_set_name():
    myGame.set_name("Aaron's Awesome Game")
    assert myGame.name == "Aaron's Awesome Game"


def test_game_set_rec_players():
    myGame.set_rec_players(2)
    assert myGame.rec_players == 2


def test_game_set_rec_tourney():
    myGame.set_rec_tourney(Single_Elimination)
    assert myGame.rec_tourney == Single_Elimination


def test_game_set_labels():
    myGame.set_labels(["Total Jokers Used"])
    assert myGame.labels == ["Total Jokers Used"]


def test_game_get_labels():
    assert myGame.get_labels() == ["Total Jokers Used"]


### INTEGRATION TESTS ###


def test_SE_custom_game_tourney():
    MyGame = Game()
    MyGame.set_name("2v2 Basketball")
    MyGame.set_labels(["Home Fouls", "Away Fouls"])

    p1 = Player("Aaron")
    p2 = Player("Xandra")
    p3 = Player("Lucas")
    p4 = Player("Tiffany")
    p5 = Player("Raf")
    p6 = Player("Rhea")
    p7 = Player("Loki")
    p8 = Player("Richard")
    p9 = Player("Owen")
    p10 = Player("Saturn")
    p11 = Player("Mo")
    p12 = Player("Sergio")
    p13 = Player("Nick")
    p14 = Player("John")
    p15 = Player("Alex")
    p16 = Player("Ben")

    Team1 = Team("t1")
    Team1.add_player(p1)
    Team1.add_player(p2)
    Team2 = Team("t2")
    Team2.add_players([p3, p4])
    Team3 = Team("t3")
    Team3.add_players([p5, p6])
    Team4 = Team("t4")
    Team4.add_players([p7, p8])
    Team5 = Team("t5")
    Team5.add_player(p9)
    Team5.add_player(p10)
    Team6 = Team("t6")
    Team6.add_players([p11, p12])
    Team7 = Team("t7")
    Team7.add_players([p13, p14])
    Team8 = Team("t8")
    Team8.add_players([p15, p16])

    Tourney2v2 = Single_Elimination(MyGame)
    Tourney2v2.add_players([Team1, Team2, Team3, Team4, Team5, Team6, Team7, Team8])
    Tourney2v2.start()
    assert Tourney2v2.matchups[1]["Home"] == "t1"
    assert Tourney2v2.matchups[1]["Away"] == "t2"
    assert Tourney2v2.matchups[1]["Next"] == 5
    assert Tourney2v2.matchups[1]["Home Fouls"] == None
    Tourney2v2.input_result(1, 21, 14, [2, 4])
    Tourney2v2.input_result(2, 21, 20, [0, 1])
    Tourney2v2.input_result(3, 18, 21, [1, 1])
    Tourney2v2.input_result(4, 10, 21, [4, 3])
    assert Tourney2v2.matchups[5]["Away"] == "t2"
    assert Tourney2v2.matchups[5]["Home"] == "t4"
    assert Tourney2v2.matchups[5]["Next"] == 7
    assert Tourney2v2.matchups[1]["Home Fouls"] == 2
    Tourney2v2.input_result(5, 21, 19, [0, 0])
    Tourney2v2.input_result(6, 18, 21, [7, 9])
    assert Tourney2v2.matchups[7]["Home"] == "t7"
    assert Tourney2v2.matchups[7]["Away"] == "t2"
    assert Tourney2v2.matchups[7]["Next"] == None
    assert Tourney2v2.matchups[6]["Away Fouls"] == 9
    Tourney2v2.input_result(7, 21, 16, [3, 2])
    assert Tourney2v2.matchups[7]["Away_Score"] > Tourney2v2.matchups[7]["Home_Score"]


####Test Classes####


class TestPlayer:
    def test_age(self):
        p1 = Player("Aaron")
        p1.set_age(22)
        assert p1.age == 22
